import logging
from typing import Any

import geopandas as gpd
import pandas as pd
import pystac

from stac_generator._types import CsvMediaType
from stac_generator.core.base.generator import VectorGenerator
from stac_generator.core.base.schema import ColumnInfo
from stac_generator.core.point.schema import CsvConfig

logger = logging.getLogger(__name__)


def read_csv(
    src_path: str,
    X_coord: str,
    Y_coord: str,
    epsg: int,
    Z_coord: str | None = None,
    T_coord: str | None = None,
    date_format: str = "ISO8601",
    columns: list[str] | list[ColumnInfo] | None = None,
) -> gpd.GeoDataFrame:
    """Read in csv from local disk

    Users must provide at the bare minimum the location of the csv, and the names of the columns to be
    treated as the X and Y coordinates. By default, will read in all columns in the csv. If columns and groupby
    columns are provided, will selectively read specified columns together with the coordinate columns (X, Y, T).

    :param src_path: path to csv file
    :type src_path: str
    :param X_coord: name of X field
    :type X_coord: str
    :param Y_coord: name of Y field
    :type Y_coord: str
    :param epsg: epsg code
    :type epsg: int
    :param Z_coord: name of Z field
    :type Z_coord: str
    :param T_coord: name of time field, defaults to None
    :type T_coord: str | None, optional
    :param date_format: format to pass to pandas to parse datetime, defaults to "ISO8601"
    :type date_format: str, optional
    :param columns: band information, defaults to None
    :type columns: list[str] | list[ColumnInfo] | None, optional
    :return: read dataframe
    :rtype: pd.DataFrame
    """
    logger.debug(f"reading csv from path: {src_path}")
    parse_dates: list[str] | bool = [T_coord] if isinstance(T_coord, str) else False
    usecols: list[str] | None = None
    # If band info is provided, only read in the required columns + the X and Y coordinates
    if columns:
        usecols = [item["name"] if isinstance(item, dict) else item for item in columns]
        usecols.extend([X_coord, Y_coord])
        if T_coord:
            usecols.append(T_coord)
        if Z_coord:
            usecols.append(Z_coord)
    df = pd.read_csv(
        filepath_or_buffer=src_path,
        usecols=usecols,
        date_format=date_format,
        parse_dates=parse_dates,
    )
    return gpd.GeoDataFrame(df, geometry=gpd.points_from_xy(df[X_coord], df[Y_coord], crs=epsg))


class PointGenerator(VectorGenerator[CsvConfig]):
    """ItemGenerator class that handles point data in csv format"""

    @staticmethod
    def create_config(source_cfg: dict[str, Any]) -> dict[str, Any]:
        if "X" not in source_cfg or "Y" not in source_cfg or "epsg" not in source_cfg:
            raise ValueError(
                f"Expects X and Y column to be described in source config for item: {source_cfg['id']}"
            )
        raw_df = read_csv(
            source_cfg["id"],
            source_cfg["X"],
            source_cfg["Y"],
            source_cfg["epsg"],
            source_cfg.get("Z"),
            source_cfg.get("T"),
            source_cfg.get("date_format", "ISO8601"),
        )
        columns = set(raw_df.columns) - {
            source_cfg["id"],
            source_cfg["X"],
            source_cfg["Y"],
            source_cfg.get("Z"),
            source_cfg.get("T"),
        }
        column_info = []
        for col in columns:
            column_info.append(ColumnInfo(name=col, description=f"{col}_description"))
        return CsvConfig(**source_cfg, column_info=column_info).model_dump(
            mode="json", exclude_none=True
        )

    def create_item_from_config(self, source_cfg: CsvConfig) -> pystac.Item:
        """Create item from source csv config

        :param source_cfg: config which contains csv metadata
        :type source_cfg: CsvConfig
        :return: stac metadata of the item described in source_cfg
        :rtype: pystac.Item
        """
        assets = {
            "data": pystac.Asset(
                href=source_cfg.location,
                description="Raw csv data",
                roles=["data"],
                media_type=CsvMediaType,
            )
        }
        raw_df = read_csv(
            source_cfg.location,
            source_cfg.X,
            source_cfg.Y,
            source_cfg.epsg,
            source_cfg.Z,
            source_cfg.T,
            source_cfg.date_format,
            source_cfg.column_info,
        )

        properties = source_cfg.model_dump(
            include={"X", "Y", "Z", "T", "column_info", "date_format", "title", "description"},
            exclude_unset=True,
            exclude_none=True,
        )
        return self.df_to_item(
            raw_df,
            assets,
            source_cfg,
            properties,
            source_cfg.epsg,
        )



import logging
import re
from typing import Any

import geopandas as gpd
import pystac
from pyproj.crs.crs import CRS

from stac_generator.core.base.generator import VectorGenerator as BaseVectorGenerator
from stac_generator.core.base.schema import ColumnInfo
from stac_generator.core.vector.schema import VectorConfig

logger = logging.getLogger(__name__)


def extract_epsg(crs: CRS) -> tuple[int, bool]:
    """Extract epsg information from crs object.
    If epsg info can be extracted directly from crs, return that value.
    Otherwise, try to convert the crs info to WKT2 and extract EPSG using regex

    Note that this method may yield unreliable result

    :param crs: crs object
    :type crs: CRS
    :return: epsg information
    :rtype: tuple[int, bool] - epsg code and reliability flag
    """
    if (result := crs.to_epsg()) is not None:
        return (result, True)
    # Handle WKT1 edge case
    wkt = crs.to_wkt()
    match = re.search(r'ID\["EPSG",(\d+)\]', wkt)
    if match:
        return (int(match.group(1)), True)
    # No match - defaults to 4326
    logger.warning(
        "Cannot determine epsg from vector file. Either provide it in the config or change the source file. Defaults to 4326 but can be incorrect."
    )
    return (4326, False)


class VectorGenerator(BaseVectorGenerator[VectorConfig]):
    """ItemGenerator class that handles vector data with common vector formats - i.e (shp, zipped shp, gpkg, geojson)"""

    @staticmethod
    def create_config(source_cfg: dict[str, Any]) -> dict[str, Any]:
        if not hasattr(source_cfg, "layers"):
            try:
                raw_df = gpd.read_file(source_cfg["location"])
            except Exception as e:
                raise ValueError(
                    "Compressed zip contains multiple vector layers. Please specify a layer in the original config"
                ) from e
        else:
            raw_df = gpd.read_file(source_cfg["location"], layer=getattr(source_cfg, "layer", None))
        columns = []
        for name in raw_df.columns:
            if name != "geometry":
                columns.append(ColumnInfo(name=name, description=f"{name}_description"))
        return VectorConfig(**source_cfg, column_info=columns).model_dump(
            mode="json", exclude_none=True
        )

    def create_item_from_config(self, source_cfg: VectorConfig) -> pystac.Item:
        """Create item from vector config

        :param source_cfg: config information
        :type source_cfg: VectorConfig
        :raises ValueError: if config epsg information is different from epsg information from vector file
        :return: stac metadata of the file described by source_cfg
        :rtype: pystac.Item
        """
        assets = {
            "data": pystac.Asset(
                href=str(source_cfg.location),
                media_type=pystac.MediaType.GEOJSON
                if source_cfg.location.endswith(".geojson")
                else "application/x-shapefile",
                roles=["data"],
                description="Raw vector data",
            )
        }
        logger.debug(f"Reading file from {source_cfg.location}")

        # Only read relevant fields
        if isinstance(source_cfg.column_info, list):
            columns = [
                col["name"] if isinstance(col, dict) else col for col in source_cfg.column_info
            ]
        else:
            columns = None
        raw_df = gpd.read_file(source_cfg.location, columns=columns, layer=source_cfg.layer)

        # Validate EPSG user-input vs extracted
        epsg, reliable = extract_epsg(raw_df.crs)
        if source_cfg.epsg is not None:
            if reliable and epsg != source_cfg.epsg:
                raise ValueError(
                    f"Source crs: {raw_df.crs} does not match config epsg: {source_cfg.epsg}"
                )
            epsg = source_cfg.epsg

        properties = source_cfg.model_dump(
            include={"column_info", "title", "description", "layer"},
            exclude_unset=True,
            exclude_none=True,
        )

        return self.df_to_item(raw_df, assets, source_cfg, properties, epsg)



from __future__ import annotations

import json
from typing import Any, cast

import pystac
import rasterio
from pyproj import CRS
from pyproj.transformer import Transformer
from pystac.extensions.eo import Band, EOExtension
from pystac.extensions.projection import ItemProjectionExtension, ProjectionExtension
from pystac.extensions.raster import AssetRasterExtension, DataType, RasterBand
from shapely import box, to_geojson

from stac_generator.core.base.generator import ItemGenerator
from stac_generator.core.base.utils import calculate_timezone

from .schema import BandInfo, RasterConfig

BAND_MAPPING: dict[str, str] = {
    "red": "red",
    "green": "green",
    "blue": "blue",
    "nir": "nir",
    "ms_red": "ms_red",
    "ms_green": "ms_green",
    "red_edge": "rededge",
    "rededge": "rededge",
    "ndvi": "ndvi",
    "ndvi2": "ndvi2",
}


class RasterGenerator(ItemGenerator[RasterConfig]):
    @staticmethod
    def create_config(source_cfg: dict[str, Any]) -> dict[str, Any]:
        with rasterio.open(source_cfg["location"]) as src:
            band_count = src.count
            bands = []
            for i in range(band_count):
                bands.append(BandInfo(name=f"Band{i}", common_name=f"Band{i}"))
        return RasterConfig(**source_cfg, band_info=bands).model_dump(
            mode="json", exclude_none=True
        )

    def create_item_from_config(self, source_cfg: RasterConfig) -> pystac.Item:
        with rasterio.open(source_cfg.location) as src:
            bounds = src.bounds
            transform = src.transform
            crs = cast(CRS, src.crs)
            shape = list(src.shape)

            # Convert to 4326 for bbox and geometry
            transformer = Transformer.from_crs(crs, 4326, always_xy=True)
            minx, miny = transformer.transform(bounds.left, bounds.bottom)
            maxx, maxy = transformer.transform(bounds.right, bounds.top)
            bbox: tuple[float, float, float, float] = (minx, miny, maxx, maxy)

            # Create geometry as Shapely Polygon
            geometry = box(*bbox)
            geometry_geojson = json.loads(to_geojson(geometry))

            # Process datetime
            item_tz = calculate_timezone(geometry)
            item_ts = source_cfg.get_datetime(item_tz)

            # Validate EPSG
            epsg = crs.to_epsg()
            if source_cfg.epsg is not None and source_cfg.epsg != epsg:
                raise ValueError(
                    f"EPSG mismatch: source_cfg.epsg ({source_cfg.epsg}) does not match EPSG ({epsg})."
                )

            # Create STAC Item
            item = pystac.Item(
                id=source_cfg.id,
                geometry=geometry_geojson,
                bbox=list(bbox),
                datetime=item_ts,
                properties={},
            )

            # Projection extension
            proj_ext = ItemProjectionExtension.ext(item, add_if_missing=True)
            affine_transform = [
                rasterio.transform.from_bounds(*bbox, shape[1], shape[0])[i] for i in range(9)
            ]
            proj_ext.apply(epsg=epsg, wkt2=crs.to_wkt(), shape=shape, transform=affine_transform)

            # Initialize extensions
            EOExtension.ext(item, add_if_missing=True)
            proj_ext = ProjectionExtension.ext(item, add_if_missing=True)
            proj_ext.apply(epsg=crs.to_epsg(), shape=shape, transform=list(transform)[:9])

            # Create EO and Raster bands
            eo_bands = []
            raster_bands = []
            for band_info in source_cfg.band_info:
                eo_band = Band.create(
                    name=band_info.name.lower(),
                    common_name=BAND_MAPPING.get(band_info.common_name.lower(), None),
                    center_wavelength=float(band_info.wavelength)
                    if band_info.wavelength is not None
                    else None,
                    description=f"Common name: {BAND_MAPPING.get(band_info.common_name.lower(), 'unknown')}",
                )
                eo_bands.append(eo_band)

                raster_band = RasterBand.create(
                    nodata=band_info.nodata or 0,
                    data_type=DataType(band_info.data_type or "uint16"),
                )
                raster_bands.append(raster_band)

            eo_ext = EOExtension.ext(item, add_if_missing=True)
            eo_ext.apply(bands=eo_bands, cloud_cover=0.0, snow_cover=0.0)

            # Create Asset and Add to Item
            asset = pystac.Asset(
                href=source_cfg.location,
                media_type=pystac.MediaType.GEOTIFF,
                roles=["data"],
                title="Raster Data",
            )
            item.add_asset("data", asset)

            # Apply Raster Extension to the Asset
            raster_ext = AssetRasterExtension.ext(asset, add_if_missing=True)
            raster_ext.apply(bands=raster_bands)

            # Add eo:bands to Asset
            # Apply EO bands to the asset using AssetEOExtension
            asset_eo_ext = EOExtension.ext(asset, add_if_missing=True)
            asset_eo_ext.apply(bands=eo_bands)

            return item



from stac_generator.core.base.schema import HasColumnInfo, SourceConfig


class VectorConfig(SourceConfig, HasColumnInfo):
    """Extended source config with EPSG code."""

    epsg: int | None = None
    """EPSG code for checking against EPSG code of the vector data"""

    layer: str | None = None
    """Vector layer for multi-layer shapefile"""
